#include <QTimer>
#include "ReceiveBuffer.h"
#include "vmc.h"
#include <QDebug>

ReceiveBuffer::ReceiveBuffer(COM_interface *cinterface, QObject *parent) :
    QObject(parent)
{    
    qRegisterMetaType<RECEIVE_ERROR>("RECEIVE_ERROR");
    this->vmc = (VMC*)parent;
    this->interface = cinterface;

    connect(interface, SIGNAL(ReadData(QByteArray*)), this, SLOT(getData(QByteArray*)));
    connect(vmc, SIGNAL(CommandProcessed()), this, SLOT(removeCommand()));
    connect(this, SIGNAL(CommandAvailiable(QByteArray)), vmc, SLOT(CommandAvaliable(QByteArray)));

    currentMessage.clear();

    receiveIdleTimer = new QTimer(this);
    receiveIdleTimer->setInterval(2000);      // clear possible wrong data in receive buffer after 2s
    receiveIdleTimer->setSingleShot(false);

    connect(receiveIdleTimer, SIGNAL(timeout()), this, SLOT(onReceiveIdleTimerTimeout()));

}


ReceiveBuffer::~ReceiveBuffer()
{
    delete(this->receiveIdleTimer);
}


int ReceiveBuffer::size()
{
    return messageQueue.size();
}


void ReceiveBuffer::removeCommand()
{
    if (messageQueue.size() > 0) {
        messageQueue.removeFirst();
    }
    else {
        qDebug() << "ReceiveBuffer::removeCommand(): no message to remove!";
    }
}


void ReceiveBuffer::onReceiveIdleTimerTimeout()
{
    this->currentMessage.clear();
    this->receiveIdleTimer->stop();
}



void ReceiveBuffer::getData(QByteArray *data)
{
    int CRCindex;
    int COMMAND_IN_PROGRESS = 1;

    // do nothing, if there is no data
    if (data->size() == 0) { return; }

    // DEBUG
    /*
          qDebug() << "ReceiveBuffer::getData(): messageQueue.size     = " << messageQueue.size();
          qDebug() << "ReceiveBuffer::getData(): data                  = " << data;
          qDebug() << "ReceiveBuffer::getData(): data->size()          = " << data->size();
          qDebug() << "ReceiveBuffer::getData(): &currentMessage       = " << &currentMessage;
          qDebug() << "ReceiveBuffer::getData(): currentMessage.capacity = " << currentMessage.capacity();
          qDebug() << "ReceiveBuffer::getData(): currentMessage.size() = " << currentMessage.size();
    */

    this->receiveIdleTimer->start();

    // for safety: prevent currentMessage to grow infinitely in size.
    if (currentMessage.size() > 0xFF) {
        emit ReceiveError(RECEIVE_ERROR::INVALID_COMMAND);
        currentMessage.clear();
    }


    // we use append because data can be an interrupted command:
    currentMessage.append(data->data(), data->size());

          //qDebug() << "ReceiveBuffer::getData(): currentMessage    = " << QString(currentMessage);


    // process commands:
    while (COMMAND_IN_PROGRESS) {

        int startindex = currentMessage.indexOf(0x3e);        // find start of command

        // detect ACK/NACK response messages from vmc:
        if (startindex != 0) {
            int forEnd = -1;
            if (startindex > 0)   forEnd = startindex;              // if ACK/NACK is before command ...
            if (startindex == -1) forEnd = currentMessage.size();   // if ACK/NACK is the only data

            for (int i = 0; i != forEnd; i++) {
                switch (currentMessage.at(i)) {
                    case ACK:  emit ReceiveResponse(ACK);
                               break;
                    case NACK: emit ReceiveResponse(NACK);
                               break;
                }
            }
        }

        if(startindex == -1) {
             currentMessage.clear();
             COMMAND_IN_PROGRESS = 0;
             return;
        }
        else {
             currentMessage.remove(0,startindex);
        }

        // now, we have at least a beginning message
        CRCindex = checkMessageForCRC();

        if (CRCindex == 0) {
           // exit this methode and message part is still in currentMessage.
           break;
        }
        if (CRCindex < 0) {

            //qCriticalBAMessage(currentMessage);

            emit ReceiveError(RECEIVE_ERROR::INVALID_COMMAND);
            currentMessage.clear();
            COMMAND_IN_PROGRESS = 0;
            break;
        }

              //qDebug() << "ReceiveBuffer::getData(): startindex     = " << startindex;
              //qDebug() << "ReceiveBuffer::getData(): currentMessage = " << currentMessage;
              //qDebug() << "ReceiveBuffer::getData(): CRCindex       = " << CRCindex;

        if (CRCindex) {
              // found a complete message
              if (checkCRC(currentMessage.at(CRCindex), CRCindex)) {
                      messageQueue << currentMessage.left(CRCindex + 1);
                      emit CommandAvailiable(messageQueue.first());
              }
              else {
                      emit ReceiveError(RECEIVE_ERROR::CRC);
              }

              if (currentMessage.size() > (CRCindex + 1)) {
                 currentMessage.remove(0,(CRCindex + 1));
              }
              else {
                 currentMessage.clear();
                 COMMAND_IN_PROGRESS = 0;
              }
        }

    }

}


int ReceiveBuffer::checkMessageForCRC()
{
    int CRCindex;
    int DataLengthIndex;


    //        qDebug() << "ReceiveBuffer::checkMessageForCRC(): currentMessage.size = " << currentMessage.size();
    //        for (int i = 0; i < currentMessage.size(); i++)
    //            qDebug() << "       currentMessage.at(" << i << ") = " << currentMessage.at(i);


    if (currentMessage.size() <= 1)
        return 0;

    switch (currentMessage.at(1)) {
        case 0x60: DataLengthIndex = 5; break;
        case 0x61: DataLengthIndex = 6; break;
        default: DataLengthIndex = 2;   break;
    }

    if (currentMessage.size() <= DataLengthIndex)
        return 0;

    if (( (static_cast<quint8>(currentMessage.at(DataLengthIndex)) - 0x30)) <= 0 )
        return -1;

    CRCindex = DataLengthIndex + ( static_cast<quint8>(currentMessage.at(DataLengthIndex)) - 0x30) + 1;

    if (currentMessage.size() <= CRCindex)
        return 0;
    else
        return CRCindex;
}



bool ReceiveBuffer::checkCRC(quint8 CRC, int index)
{
    quint8 cCRC = 0;

    for (int i = 0; i < index; i++) {
        cCRC += static_cast<quint8>(currentMessage.at(i));
        //qDebug() << "ReceiveBuffer::checkCRC(): currentMessage.at(i) = " << currentMessage.at(i);
    }

    //qDebug() << "ReceiveBuffer::checkCRC(): cCRC = " << cCRC;
    //qDebug() << "ReceiveBuffer::checkCRC():  CRC = " <<  CRC;


    if (cCRC == CRC) return true;
               else  return false;
}



/**
 * @brief ReceiveBuffer::qCriticalBAMessage
 * @param message
 *
 * This is for test and debugging only.
 */
void ReceiveBuffer::qCriticalBAMessage(QByteArray message)
{
    QString hexString = "received: ";
    qint64 i;

    for (i=0; i < message.size(); i++) {
        hexString.append((QString(message.at(i))).toLatin1().toHex());
        hexString.append(' ');
    }
    hexString.chop(1);

    //qCritical() << "VMC::CommandAvaliable(): message.size(): " << message.size() << "";
    qCritical() << "VMC::qCriticalBAMessage(): message: "        << hexString      << "";

}


